from datetime import datetime, timezone

import pytest

import adif_io


def test_adif_str_import() -> None:
    """Plain vanilla import test."""
    qsos, header = adif_io.read_from_string(
        "A sample ADIF content for demonstration.\n"
        "<adif_ver:5>3.1.3<eoh>\n"
        "<QSO_DATE:8>20190714 <TIME_ON:4>1140<CALL:5>LY0HQ"
        "<MODE:2>CW<BAND:3>40M<RST_SENT:3>599<RST_RCVD:3>599"
        "<STX_STRING:2>28<SRX_STRING:4>LRMD<EOR>\n"
        "<QSO_DATE:8>20190714<TIME_ON:4>1130<CALL:5>SE9HQ<MODE:2>CW<FREQ:1>7"
        "<BAND:3>40M<RST_SENT:3>599<RST_RCVD:3>599"
        "<SRX_STRING:3>SSA<DXCC:3>284<EOR>"
    )

    assert [
        {
            "QSO_DATE": "20190714",
            "TIME_ON": "1140",
            "CALL": "LY0HQ",
            "MODE": "CW",
            "BAND": "40M",
            "RST_SENT": "599",
            "RST_RCVD": "599",
            "STX_STRING": "28",
            "SRX_STRING": "LRMD",
        },
        {
            "QSO_DATE": "20190714",
            "TIME_ON": "1130",
            "CALL": "SE9HQ",
            "MODE": "CW",
            "FREQ": "7",
            "BAND": "40M",
            "RST_SENT": "599",
            "RST_RCVD": "599",
            "SRX_STRING": "SSA",
            "DXCC": "284",
        },
    ] == qsos

    assert {"ADIF_VER": "3.1.3"} == header
    assert datetime(
        year=2019, month=7, day=14, hour=11, minute=40, second=0, tzinfo=timezone.utc
    ) == adif_io.time_on(qsos[0])
    assert datetime(
        year=2019, month=7, day=14, hour=11, minute=30, second=0, tzinfo=timezone.utc
    ) == adif_io.time_on(qsos[1])

    with pytest.raises(KeyError) as exinfo:
        adif_io.time_off(qsos[0])
    exinfo.match("TIME_OFF")


def test_adif_file_import() -> None:
    """Plain vanilla import test."""
    qsos, header = adif_io.read_from_file("tests/vanilla.adi")

    assert [
        {
            "QSO_DATE": "20190714",
            "TIME_ON": "1140",
            "CALL": "LY0HQ",
            "MODE": "CW",
            "BAND": "40M",
            "RST_SENT": "599",
            "RST_RCVD": "599",
            "STX_STRING": "28",
            "SRX_STRING": "LRMD",
        },
        {
            "QSO_DATE": "20190714",
            "TIME_ON": "113027",
            "CALL": "SE9HQ",
            "MODE": "CW",
            "FREQ": "7",
            "BAND": "40M",
            "RST_SENT": "599",
            "RST_RCVD": "599",
            "SRX_STRING": "SSA",
            "DXCC": "284",
        },
    ] == qsos

    assert {"ADIF_VER": "3.1.3"} == header


def test_convert_field_names_to_upper_case() -> None:
    qsos, header = adif_io.read_from_string(
        "<QSo_DaTE:8>20190714 <time_on:4>1140<Call:5>LY0HQ"
        "<mODE:2>CW<band:3>40M<RSt_SENt:3>599<RST_RCVD:3>599"
        "<STX_STRING:2>28<SRX_STRING:4>LRMD<EOR>\n"
        "<QSO_DATE:8>20190714<TIME_ON:4>1130<CALL:5>SE9HQ<MODE:2>CW<FREQ:1>7"
        "<BAND:3>40M<RST_SENT:3>599<RST_RCVD:3>599"
        "Let us see that the example from the README also works:"
        "<notes:66>In this QSO, we discussed ADIF and in particular the <eor> marker."
        "<SRX_STRING:3>SSA<DXCC:3>284<Eor>"
    )

    assert [
        {
            "QSO_DATE": "20190714",
            "TIME_ON": "1140",
            "CALL": "LY0HQ",
            "MODE": "CW",
            "BAND": "40M",
            "RST_SENT": "599",
            "RST_RCVD": "599",
            "STX_STRING": "28",
            "SRX_STRING": "LRMD",
        },
        {
            "QSO_DATE": "20190714",
            "TIME_ON": "1130",
            "CALL": "SE9HQ",
            "MODE": "CW",
            "FREQ": "7",
            "BAND": "40M",
            "RST_SENT": "599",
            "RST_RCVD": "599",
            "SRX_STRING": "SSA",
            "DXCC": "284",
            "NOTES": "In this QSO, we discussed ADIF and in particular the <eor> marker.",
        },
    ] == qsos

    assert {} == header


def test_need_eoh_after_header() -> None:
    with pytest.raises(adif_io.AdifHeaderWithoutEOHError):
        qsos, header = adif_io.read_from_string("Some header <qso_date:8>20230606<eor>")


def test_double_header_field_rejected() -> None:
    with pytest.raises(adif_io.AdifDuplicateFieldError) as exinfo:
        qsos, header = adif_io.read_from_string(
            "A sample ADIF content for demonstration.\n"
            "<adif_ver:5>3.1.3<adif_ver:5>3.0.2<eoh>\n"
            "<QSO_DATE:8>20190714 <TIME_ON:4>1140<CALL:5>LY0HQ"
            "<MODE:2>CW<BAND:3>40M<RST_SENT:3>599<RST_RCVD:3>599"
            "<STX_STRING:2>28<SRX_STRING:4>LRMD<EOR>\n"
            "<QSO_DATE:8>20190714<TIME_ON:4>1130<CALL:5>SE9HQ<MODE:2>CW<FREQ:1>7"
            "<BAND:3>40M<RST_SENT:3>599<RST_RCVD:3>599"
            "<SRX_STRING:3>SSA<DXCC:3>284"
            "<qso_date:8>20230606<EOR>"
        )
    exinfo.match(r"3\.1\.3")
    exinfo.match(r"3\.0\.2")
    exinfo.match("ADIF_VER")


def test_double_qso_field_rejected() -> None:
    with pytest.raises(adif_io.AdifDuplicateFieldError) as exinfo:
        adif_io.read_from_string(
            "A sample ADIF content for demonstration.\n"
            "<adif_ver:5>3.1.3<eoh>\n"
            "<QSO_DATE:8>20190714 <TIME_ON:4>1140<CALL:5>LY0HQ"
            "<MODE:2>CW<BAND:3>40M<RST_SENT:3>599<RST_RCVD:3>599"
            "<STX_STRING:2>28<SRX_STRING:4>LRMD<EOR>\n"
            "<QSO_DATE:8>20190714<TIME_ON:4>1130<CALL:5>SE9HQ<MODE:2>CW<FREQ:1>7"
            "<BAND:3>40M<RST_SENT:3>599<RST_RCVD:3>599"
            "<SRX_STRING:3>SSA<DXCC:3>284"
            "<qso_date:8>20230606<EOR>"
        )

    exinfo.match("20190714")
    exinfo.match("20230606")
    exinfo.match("QSO_DATE")


def test_time_off_date_heuristic() -> None:
    qsos, header = adif_io.read_from_string(
        "A sample ADIF content for demonstration.\n"
        "<adif_ver:5>3.1.3<eoh>\n"
        "<QSO_DATE:8>20190714 <TIME_ON:6>114018 <TIME_OFF:4>2358 <EOR>\n"
        "<QSO_DATE:8>20190714 <TIME_ON:6>115019 <TIME_OFF:4>1023 <EOR>\n"
        "<QSO_DATE:8>20190714 <TIME_ON:6>115020 <QSO_DATE_OFF:8>20190720 <TIME_OFF:6>115152 <EOR>\n"
    )

    assert datetime(
        year=2019, month=7, day=14, hour=11, minute=40, second=18, tzinfo=timezone.utc
    ) == adif_io.time_on(qsos[0])
    assert datetime(
        year=2019, month=7, day=14, hour=23, minute=58, tzinfo=timezone.utc
    ) == adif_io.time_off(qsos[0])

    assert datetime(
        year=2019, month=7, day=14, hour=11, minute=50, second=19, tzinfo=timezone.utc
    ) == adif_io.time_on(qsos[1])
    assert datetime(
        year=2019, month=7, day=15, hour=10, minute=23, tzinfo=timezone.utc
    ) == adif_io.time_off(qsos[1])

    assert datetime(
        year=2019, month=7, day=14, hour=11, minute=50, second=20, tzinfo=timezone.utc
    ) == adif_io.time_on(qsos[2])
    assert datetime(
        year=2019, month=7, day=20, hour=11, minute=51, second=52, tzinfo=timezone.utc
    ) == adif_io.time_off(qsos[2])



from pytest import approx

import adif_io


def test_plain_vanilla() -> None:
    assert approx(52 + 26.592 / 60) == adif_io.degrees_from_location("N052 26.592")
    assert "N052 26.592" == adif_io.location_from_degrees(52 + 26.592 / 60, True)

    assert approx(-(17 + 3.123 / 60)) == adif_io.degrees_from_location("S017 03.123")
    assert "S017 03.123" == adif_io.location_from_degrees(-(17 + 3.123 / 60), True)

    assert approx(0.001 / 60) == adif_io.degrees_from_location("E000 00.001")
    assert "E000 00.001" == adif_io.location_from_degrees(0.001 / 60, False)

    assert approx(-90) == adif_io.degrees_from_location("W090 00.000")
    assert "W090 00.000" == adif_io.location_from_degrees(-90, False)

    assert approx(179.5) == adif_io.degrees_from_location("E179 30.000")
    assert "E179 30.000" == adif_io.location_from_degrees(179.5, False)



import setuptools

with open("README.md", "r", encoding="utf_8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="adif_io",
    version="0.4.0",
    author="Andreas Krüger (DJ3EI)",
    author_email="dj3ei@famsik.de",
    description="Basic input of ADIF radio amateur log files.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    url="https://gitlab.com/andreas_krueger_py/adif_io",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: Apache Software License",
        "Topic :: Communications :: Ham Radio"
    ],
    python_requires='>=3.6',
)



#  Copyright 2019, 2023 Andreas Krüger, DJ3EI
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# This is an ADIF parser in Python.

# It knows nothing about ADIF data types or enumerations,
# everything is a string, so it is fairly simple.

# But it does correcly handle things like:
# <notes:66>In this QSO, we discussed ADIF and in particular the <eor> marker.
# So, in that sense, this parser is somewhat sophisticated.

# Main result of parsing: List of QSOs.
# Each QSO is one Python dict.
# Keys in that dict are ADIF field names in upper case,
# value for a key is whatever was found in the ADIF, as a string.
# Order of QSOs in the list is same as in ADIF file.

import math
import re
from datetime import datetime, timedelta, timezone


class AdifError(Exception):
    """Base error."""

    pass


class AdifHeaderWithoutEOHError(AdifError):
    """Error for header found, but not terminated with <EOH>"""

    pass


class AdifDuplicateFieldError(AdifError):
    """Error for duplicate fileds in one QSO record or in the header."""

    pass


# Some type definitions:

headers = dict[str, str]
qso = dict[str, str]


def read_from_string(adif_string: str) -> tuple[list[qso], headers]:
    # The ADIF file header keys and values, if any.
    adif_headers: headers = {}

    header_field_re = re.compile(r"<((eoh)|(\w+)\:(\d+)(\:[^>]+)?)>", re.IGNORECASE)
    field_re = re.compile(r"<((eor)|(\w+)\:(\d+)(\:[^>]+)?)>", re.IGNORECASE)

    qsos: list[qso] = []
    cursor = 0
    if adif_string[0] != "<":
        # Input has ADIF header. Read all header fields.
        eoh_found = False
        while not eoh_found:
            header_field_mo = header_field_re.search(adif_string, cursor)
            if header_field_mo:
                if header_field_mo.group(2):
                    eoh_found = True
                    cursor = header_field_mo.end(0)
                else:
                    field = header_field_mo.group(3).upper()
                    value_start = header_field_mo.end(0)
                    value_end = value_start + int(header_field_mo.group(4))
                    value = adif_string[value_start:value_end]
                    if field in adif_headers:
                        raise AdifDuplicateFieldError(
                            f'Duplication in ADI header, {field} previously "{adif_headers[field]}", now "{value}".'
                        )
                    adif_headers[field] = value
                    cursor = value_end
            else:
                raise AdifHeaderWithoutEOHError(
                    "<EOF> marker missing after ADIF header."
                )

    one_qso: qso = {}
    field_mo = field_re.search(adif_string, cursor)
    while field_mo:
        if field_mo.group(2):
            # <eor> found:
            qsos.append(one_qso)
            one_qso = {}
            cursor = field_mo.end(0)
        else:
            # Field found:
            field = field_mo.group(3).upper()
            value_start = field_mo.end(0)
            value_end = value_start + int(field_mo.group(4))
            value = adif_string[value_start:value_end]
            if field in one_qso:
                raise AdifDuplicateFieldError(
                    f'Duplication in qso {qso}, {field} previously "{one_qso[field]}", now "{value}".'
                )
            one_qso[field] = value
            cursor = value_end
        field_mo = field_re.search(adif_string, cursor)

    return (qsos, adif_headers)


def read_from_file(filename: str) -> tuple[list[qso], headers]:
    with open(filename) as adif_file:
        adif_string = adif_file.read()
        return read_from_string(adif_string)


_ONE_DAY = timedelta(days=1)


def time_on(one_qso: qso) -> datetime:
    date = one_qso["QSO_DATE"]
    y = int(date[0:4])
    mo = int(date[4:6])
    d = int(date[6:8])
    time = one_qso["TIME_ON"]
    h = int(time[0:2])
    mi = int(time[2:4])
    s = int(time[4:6]) if len(time) == 6 else 0
    return datetime(y, mo, d, h, mi, s, tzinfo=timezone.utc)


def time_off(one_qso: qso) -> datetime:
    if "QSO_DATE_OFF" in one_qso:
        date = one_qso["QSO_DATE_OFF"]
        y = int(date[0:4])
        mo = int(date[4:6])
        d = int(date[6:8])
        time = one_qso["TIME_OFF"]
        h = int(time[0:2])
        mi = int(time[2:4])
        s = int(time[4:6]) if len(time) == 6 else 0
        return datetime(y, mo, d, h, mi, s, tzinfo=timezone.utc)
    else:
        date = one_qso["QSO_DATE"]
        y = int(date[0:4])
        mo = int(date[4:6])
        d = int(date[6:8])
        time = one_qso["TIME_OFF"]
        h = int(time[0:2])
        mi = int(time[2:4])
        s = int(time[4:6]) if len(time) == 6 else 0
        time_off_maybe = datetime(y, mo, d, h, mi, s, tzinfo=timezone.utc)
        if time_on(one_qso) < time_off_maybe:
            return time_off_maybe
        else:
            return time_off_maybe + _ONE_DAY


def degrees_from_location(adif: str) -> float:
    """Convert an ADIF location string to degrees."""
    x = adif[0]
    deg_i = int(adif[1:4])
    min = float(adif[5:])
    deg = deg_i + min / 60
    return deg if x in ["N", "E", "n", "e"] else -deg


def location_from_degrees(degrees: float, lat: bool) -> str:
    """Convert degrees to an ADIF location string, either latitude or longitude.

    If the `lat` parameter is true, N / S are used,
    if false, E / W.
    """
    if lat:
        if degrees < 0.0:
            x = "S"
        else:
            x = "N"
    else:
        if degrees < 0.0:
            x = "W"
        else:
            x = "E"
    degrees_abs = abs(degrees)
    deg_num = int(math.floor(degrees_abs))
    min_num = (degrees_abs - deg_num) * 60
    return f"{x}{deg_num:03d} {min_num:06.3f}"
